// -----------------------------------------------------------------------
// file:        grdump.cpp
// version:     0.5d (October 22, 2005)
//
// description: This is a very simple example using the msxgr.cpp api
//              to use the MSX Game Reader. This dumper works more or
//              less the same as the saverom utility on a real MSX.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2, or (at your option)
//  any later version. See COPYING for more details.
//
// Copyright 2005 Vincent van Dam (vincentd@erg.verweg.com)
// -----------------------------------------------------------------------

#include <stdio.h>

#include "msxgr.h"

// Display an error message and exit programm

void ExitError(char *Message)
{
	printf("%s\n",Message);
	exit(1);
}

// Display a list of connected MSX Game Readers and exit.

void DisplayListReaders()
{
	CMSXGr *msxgr = new CMSXGr();
	if (msxgr->Init()) {
		printf("failed to init gamereader\n");
		exit(1);
	}
	
	bool FoundReader = false;
	for (int Slot=0;Slot<16;Slot++)
		if (msxgr->IsSlotEnable(Slot)) {
			FoundReader = true;
			bool CartridgeInserted = msxgr->IsCartridgeInserted(Slot);	
			printf( "Found MSX Game Reader, Slot %d, %s.\n", Slot,
				CartridgeInserted ? "cartridge inserted"
						  : "no cartridge inserted" );
		}
		
	if (!FoundReader) 
		ExitError("MSX GameReader not found!");
		
	msxgr->Uninit();
	return;
}

// Dump ROM image for given slot and exit.

void DumpRomImage(int Slot,char* OutputFilename,int MapperAddress,int SaveAddress,int BlockSize,int Blocks)
{
	CMSXGr *msxgr = new CMSXGr();
	if (msxgr->Init()) {
		printf("failed to init gamereader\n");
		exit(1);
	}
	
	if (Slot==-1)
		for (Slot=0;Slot<17;Slot++) {
			if (Slot==16)
				ExitError("MSX Game Reader not found!");
			if (msxgr->IsSlotEnable(Slot))
				break;
		}
		
	if (!msxgr->IsCartridgeInserted(Slot))
		ExitError("No ROM inserted in MSX Game Reader!");
		
	FILE *OutputFile = fopen(OutputFilename,"wb+");
	if(!OutputFile)
		ExitError("Error writing ROM file!");
		
	printf("Dumping ROM in slot %d",Slot);
	for (int Block=0;Block<Blocks;Block++) {
		printf(".");
		// write to mapper port
		if (Blocks>1) {
			char *WriteBuffer = (char*)GlobalAlloc(GPTR,1);
			*WriteBuffer = Block;
			if (msxgr->WriteMemory(Slot,WriteBuffer,MapperAddress,1)) {
				printf("\nMSX Game Reader error: ");
				ExitError(msxgr->GetLastErrorStr());
			}
			GlobalFree(WriteBuffer);
		}
		// read data from cartridge
		char *ReadBuffer = (char*)GlobalAlloc(GPTR,BlockSize);
		if (msxgr->ReadMemory(Slot,ReadBuffer,SaveAddress,BlockSize)) {
			printf("\nMSX Game Reader error: ");
			ExitError(msxgr->GetLastErrorStr());
		}
		// write data to file
		fwrite(ReadBuffer,1,BlockSize,OutputFile);
		GlobalFree(ReadBuffer);
	}
	fclose(OutputFile);
	printf("done\n");
		
	msxgr->Uninit();
	return;
}

// Convert hex string to integer

int axtoi(char *hex)
{
	int value=0;
	for(int i=0;hex[i]!=0;i++) {
		value *= 16;
		if(hex[i]>='A' && hex[i]<='F')
			value += hex[i]-'A'+10;
		if(hex[i]>='a' && hex[i]<='f')
			value += hex[i]-'a'+10;
		if(hex[i]>='0' && hex[i]<='9')
			value += hex[i]-'0';
	}
	return value;
}

// Let's do it!

void main(int argc, char* argv[])
{
	int   i,j;
	bool  ListReaders = false;
	int   Slot = -1;
	char* OutputFilename = NULL;
	int   MapperAddress = 0x4000;
	int   SaveAddress = 0x4000;
	int   BlockSize = 0x4000;
	int   Blocks = 1;
	
	
	printf("MSX Game Reader rom dumper v0.5d - (C) 2005 Vincent van Dam\n");
	printf("-----------------------------------------------------------\n");
	
	// Parse command line options
	for (i=1; i<argc; i++) {
		if (argv[i][0]=='-') {
			for(j=1;j && argv[i][j]!='\0';j++)
				switch(argv[i][j]) {
					case 'l':
						ListReaders = true;
						break;
					case 'i':
						Slot = atoi(argv[++i]);
						j=-1;
						break;
					case 'm':
						MapperAddress = axtoi(argv[++i]);
						j=-1;
						break;
					case 'n':
						Blocks = atoi(argv[++i]);
						j=-1;
						break;
					case 's':
						BlockSize = axtoi(argv[++i]);
						j=-1;
						break;
					case 'b':
						SaveAddress = axtoi(argv[++i]);
						j=-1;
						break;
					default:
						fprintf(stderr,"grdump: invalid option\n");
						exit(1);
			}
			continue;
		}
		if (OutputFilename==NULL) {
			OutputFilename=argv[i];
			continue;
		}
		fprintf(stderr,"grdump: invalid option\n");
		exit(1);
	}
	
	// Dispatch
	if (ListReaders) {
		DisplayListReaders();
		exit(0);
	}
	
	if (OutputFilename) {
		DumpRomImage(Slot,OutputFilename,MapperAddress,SaveAddress,BlockSize,Blocks);
		exit(0);
	}
	
	// Show information and exit
	printf("Syntax: grdump <parameters> <filename>\n");
	printf("\n");
	printf("Parameters:\n");
	printf("  -l ............ list available MSX Game Readers.\n");
	printf("  -i <slot> ..... use given slot for dumping cartridge (optional).\n");
	printf("  -b <address> .. begin address of block (hex).\n");
	printf("  -s <size> ..... size of block (hex).\n");
	printf("  -m <address> .. address of rom mapper (hex).\n");
	printf("  -n <blocks> ... number of blocks to dump.\n");
	printf("  filename ...... filename to dump rom image to.\n");
	printf("\n");
	printf("Examples:\n");
	printf("\n");
	printf("  List available MSX Game Readers and their status;\n");
	printf("   grdump -l\n");
	printf("\n");
	printf("  Dump a 128 kB SCC ROM;\n");
	printf("   grdump -b a000 -s 2000 -m b000 -n 16 konami1mb.rom\n");
	printf("\n");
	printf("  Dump a 32 kB ROM;\n");
	printf("   grdump -b 4000 -s 8000 normal32kb.rom\n");
	exit(1);
}
